/*****************************************************************************
 * Copyright © 2015, SAS Institute Inc., Cary, NC, USA.  All Rights Reserved.
 ****************************************************************************/

import groovy.io.FileType
import groovy.json.JsonSlurper
import groovy.transform.Memoized

import org.codehaus.groovy.control.CompilationFailedException

import java.net.InetAddress
import java.net.UnknownHostException
import java.nio.file.Files

import org.apache.commons.logging.LogFactory
import org.apache.commons.logging.Log
import org.apache.http.conn.HttpHostConnectException

// Policy for handling file ownership issues
enum UserPolicy { IGNORE, WARN, ERROR }

/**
 * Provides programmatic access to validate Service Architecture at different levels
 * of granularity.
 * <ul>
 *   <li>Level 1 - Local environment (classpath, required directories, etc.)</li>
 *   <li>Level 2 - External systems - SAS execution, connection to SASEV Server</li>
 *   <li>Level 3 - Reporting properties, success of enabled ETLs</li>
 *   <li>Level 4 - File checksums, generate diagnostics file</li>
 * </ul>
 */
class Validator
{
    Log out
    File outFile
    EMIBaseScript emi
    UserPolicy userPolicy = UserPolicy.ERROR
    boolean includeSlowChecks = false
    boolean verbose = false
    String runID

    public Validator(EMIBaseScript base, Map flags=[:], Log outLog=null) {
        if (flags["silent"]) {
            out = null
        } else {
            out = outLog ?: LogFactory.getLog(Validator.class)
            verbose = flags["verbose"] ?: false
        }
        includeSlowChecks = flags["includeSlowChecks"]

        emi = base
        runID = emi._LOG_TIMESTAMP

        if (flags["outFile"]) {
            outFile = emi.canonicalFile(flags["outFile"]?.toString())

            if (outFile?.exists()) {
                if (flags["overwrite"]) {
                    if (!outFile.delete()) {
                        out?.error "Unable to delete existing output file: $outFile"
                        outFile = null
                    }
                } else {
                    out?.error "The output file [$outFile] exists and [overwrite] option was not specified"
                    outFile = null
                }
            }
        }
        if (flags["userPolicy"]) {
            userPolicy = UserPolicy.valueOf(flags["userPolicy"])
            if (verbose) { out?.info "Validate user policy override: $userPolicy" }
        } else if (emi.env["EMI_USER_POLICY"]) {
            userPolicy = UserPolicy.valueOf(emi.env["EMI_USER_POLICY"])
            if (verbose) { out?.info "Validate user policy environment override: $userPolicy" }
        } else if (emi.sysProps['custom.emi.user.policy']) {
            userPolicy = UserPolicy.valueOf(sysProps['custom.emi.user.policy'])
            if (verbose) { out?.info "Validate user policy property override: $userPolicy" }
        }
    }

    // Validates at the given level (does not run cumulatively)
    public ValidateResults validate(int level) {
        out?.debug "Service Architecture Validator - ${new Date()}"
        ValidateResults results = new ValidateResults()

        switch (level) {
            case 1:
               results << validateLevel1()
               break
            case 2:
               results << validateLevel2()
               break
            case 3:
               results << validateLevel3()
               break
           case 4:
               results << validateLevel4()
               break
            default:
                results.errors << "Invalid validation level specified: $level"
        }

        return results
    }

    private ValidateResults validateLevel1() {
        ValidateResults results = new ValidateResults()

        out?.info "Starting level 1 validation..."
        long start = System.currentTimeMillis()

        // Check Service Architecture System Property values
        List keyProps = [
            'sas.levroot', 'sasev.installDir', 'sasev.configDir',
            'sasev.dir', 'sasev.agent.dir',
            'sasev.landingZoneDir', 'sasev.archiveDir', 'sasev.datamartDir',
            'sas.home', 'sas.root', 'sas.exe'
        ].sort()
        keyProps.each { key ->
            if (verbose) { out?.info "\tValidating system property [$key]..." }
            if (!emi.sysProps[key]) {
                results.errors << "The required property $key is missing"
            } else if (!new File(emi.sysProps[key]).absolute) {
                results.warnings << "$key is a relative path that should be absolute (value=${emi.sysProps[key]})"
            } else if (!new File(emi.sysProps[key]).exists()) {
                results.errors << "Missing expected directory: ${emi.sysProps[key]}"
            }
        }

        if (verbose) { out?.info "Completed validation of ${keyProps.size()} Service Architecture system properties" }

        if (verbose) { out?.info "Validating sasev.properties..." }
        File sasevPropFile = new File(emi.frameworkDir, "sasev.properties")
        if (sasevPropFile.exists()) {
            try {
                Properties p = emi.readProperties(sasevPropFile)
                if (verbose) {
                    out?.info "\tFramework initialized = ${p['framework.initialized']}"
                    out?.info "\tACM ETL Enabled       = ${p['acm.enabled']}"
                    out?.info "\tAPM ETL Enabled       = ${p['apm.enabled']}"
                }
            } catch (Exception e) {
                if (verbose) { out?.error("Could not read sasev.properties", e) }
                e.printStackTrace()
                results.errors << "Could not read sasev.properties : $e"
            }
        } else {
            results.errors << "$sasevPropFile does not exist"
        }

        // Check that upgrade in place is not pending
        if (emi.sysProps['ugip.status'] == 'pending') {
            results.errors << "Upgrade-in-place has not completed successfully. Run the ugip_init command."
        }

        // Check current user vs. install user
        if (!emi.isWindows() && !emi.installUser) {
            if (userPolicy == UserPolicy.ERROR) {
                results.errors << "The current user [${emi.sysProps['user.name']}], is not the SAS install user [$emi.installUsername]"
            } else if (userPolicy == UserPolicy.WARN) {
                results.warnings << "WARNING: The current user [${emi.sysProps['user.name']}], is not the SAS install user [$emi.installUsername]"
            } else { // UserPolicy.IGNORE
                out?.debug "\tThe current user [${emi.sysProps['user.name']}], is not the SAS install user [$emi.installUsername], but the validate policy is set to IGNORE"
            }
        }

        out?.info "Level 1 validation completed with ${results.errors.size()} error(s)"

        return results
    }

    private ValidateResults validateLevel2() {
        ValidateResults results = new ValidateResults()

        if (includeSlowChecks) {
            out?.info "Starting extended level 2 validation (this could take a while)..."
        } else {
            out?.info "Starting level 2 validation..."
        }

        long start = System.currentTimeMillis()

        if (verbose) { out?.info "Checking that key resources have been discovered..." }
        List keyResourceTypes = [
            'SAS Environment Manager Data Mart 9.4',
            'SAS System Info'
        ]
        keyResourceTypes.each { resType ->
            if (!emi.hq.getResources(resType)) {
                results.errors << "No [$resType] resources found"
            }
        }

        // If slow checks are enabled, validate that the SAS Environment
        // Manager Server and Agents can handle resource updates in a reasonable
        // amount of time (can be slow immediately after server initialization)
        long maxWaitTimeSec = Long.parseLong(emi.sysProps['custom.validate.slowCheckTimeout'] ?: "45")
        boolean slowChecksComplete = false
        boolean updateTestPassed = false
        if (includeSlowChecks) {
            if (verbose) {
                out?.info "Checking if SAS Environment Manager is able to process updates..."
            }
            List cmd = [
                emi.canonicalPath(emi.emiBaseDir, "Scripts/RunGroovy${emi.windows? '.exe':''}"),
                emi.canonicalPath(emi.emiBaseDir, 'Scripts/sasevValidate.groovy')
            ]
            if (verbose) { cmd << '-v' }
            if (out?.debugEnabled) {
                out?.debug "Extended validation command to execute:"
                cmd.each {
                    out?.debug "\t$it"
                }
            }

            emi.exec cmd, async:true, asyncTimeOut:maxWaitTimeSec, workDir:emi.emiBaseDir, { rc, stdout, stderr ->
                if (rc != 0) {
                    out?.debug "Update test failed with RC $rc. Adding error to validation results."
                    out?.debug "\tstderr: $stdout"
                    out?.debug "\tstderr: $stderr"
                } else {
                    if (verbose) { out?.info "SAS Environment Manager Server is ready for resource updates" }
                    updateTestPassed = true
                }
                slowChecksComplete = true
            }
        } else {
            slowChecksComplete = true
            updateTestPassed = true
        }
        // Wait for process to complete or timeout
        long hqTestStart = System.currentTimeMillis()
        long waitTime = 0
        long lastWaitMessage = hqTestStart
        if (verbose) { out?.info "\tWaiting up to $maxWaitTimeSec seconds..." }
        while (!slowChecksComplete && waitTime < maxWaitTimeSec*1000) {
            long now = System.currentTimeMillis()
            Thread.sleep(200)
            waitTime = now - hqTestStart
            if (verbose && (System.currentTimeMillis() - lastWaitMessage) > 4000) {
                out?.info "\t[${waitTime/1000.0} sec] Waiting for update to complete..."
                lastWaitMessage = now
            }
        }
        if (!updateTestPassed) {
            results.errors << "The SAS Environment Manager Server is unable to process " +
                              "additional tasks at this time. This may be due to high " +
                              "change request volume. If the server has recently been " +
                              "(re)started, it may be synchronizing with its Agents. " +
                              "This process could take up to an hour.\n" +
                              "Please try again at a later time to validate your environment."
            // Need to stop here. Otherwise, we will hang on upcoming
            // HQApi calls
            out?.debug "Skipping remaining level 2 checks to avoid blocking on further HQApi calls"
            return results
        }

        // Check the SAS Environment Manager Agent
        if (verbose) { out?.info "Checking SAS Environment Manager Agent status..." }
        if (!emi.hq.agentAvailable) {
            if (emi.hq.agentRunning) {
                results.errors << "The SAS Environment Manager Agent is running but is " +
                "not properly reporting monitoring data to the server.  Check the agent.log file."
            } else {
                results.errors << "The SAS Environment Manager Agent is not running."
            }
        }

        // Check emi-framework SAS Environment Manager plugins have been loaded and that
        // key resources have been created
        try {
            if (verbose) { out?.info "Checking for SAS Environment Manager Server connectivity..." }
            if (!emi.hq.api.resourceApi.getResourcePrototype("HQ Agent")) {
                results.errors << "Unable to connect to the SAS Environment Manager Server"
            } else {
                // SAS Environment Manager Server connectivity is OK
                if (verbose) { out?.info "Checking for SAS Metadata Server 9.4 resource type..." }
                if (!emi.hq.api.resourceApi.getResourcePrototype("SAS Metadata Server 9.4")) {
                    results.errors << "SAS plugins were not found in SAS Environment Manager"
                }

                if (verbose) { out?.info "Checking for SAS Environment Manager Data Mart 9.4 resource type..." }
                if (!emi.hq.api.resourceApi.getResourcePrototype("SAS Environment Manager Data Mart 9.4")) {
                    results.errors << "SAS emi-framework plugin is not loaded in SAS Environment Manager"
                }

                // Check if SASEV User is a Super User
                String connUser = emi.hq.connectionInfo["user"]
                if (verbose) { out?.info "Checking SAS Environment Manager user permissions for $connUser..." }
                boolean superUser = false
                def roleResp = emi.hq.api.roleApi.roles
                if (roleResp) {
                    roleResp.role.each { role ->
                        if (role.name == "Super User Role") {
                            role.user.each { user ->
                                out?.debug "Checking Super User Role member: [$user.name] against [$connUser]"
                                // Also check '@saspw' suffix
                                if (user.name == connUser || "${user.name}@saspw" == connUser) {
                                    superUser = true
                                    out?.debug "The user $user.name has the Super User Role"
                                    return // from closure
                                }
                            }
                        }
                    }
                }
                if (!superUser) {
                    results.errors << "$connUser is not a member of the Super Users Role"
                }

                // Find the local agent
                if (verbose) { out?.info "Checking SAS Environment Manager Agent status..." }
                boolean foundAgent = false
                def agentsResp = emi.hq.api.agentApi.agents
                if (agentsResp) {
                    agentsResp.agent.each { agent ->
                        if (emi.net.isLocal(agent.address)) {
                            if (verbose) { out?.info "Found local SAS Environment Manager Agent registered at: $agent.address" }
                            foundAgent = true
                            return // from closure
                        }
                    }
                }
                if (!foundAgent) {
                    File agentLogFile = emi.canonicalFile(emi.sysProps['sasev.agent.dir'], "log/agent.log")
                    results.errors << "The local SAS Environment Agent is not properly initialized. " +
                              "Check the agent log file at $agentLogFile. Restarting the agent may resolve the issue."
                }
            }
        } catch (HttpHostConnectException httpEx) {
            if (verbose) { out?.error "Unable to connect to SAS Environment Manager Server: $httpEx.message" }
            results.errors << "Unable to connect to SAS Environment Manager Server: $httpEx.message"
        } catch (Exception e) {
            if (verbose) { out?.error "There was a problem querying the SAS Environment Manager Server: $e" }
            results.errors << "There was a problem querying the SAS Environment Manager Server: $e"
        }

        // quick check if a stored process server has been configured
        if (verbose) { out?.info "Checking for a Stored Process Server..." }
        def stpResources = emi.hq.getResources("SAS Object Spawner 9.4 SAS Stored Process Server")
        if (stpResources?.size() == 0) {
            out?.error "No configured Stored Process Server found"
            results.errors << "No configured Stored Process Server found"
        }

        // Check that all .groovy files compile properly
        if (verbose) { out?.info "Checking that all Groovy scripts parse properly..." }
        emi.canonicalFile(emi.sysProps["emi.framework.scripts.dir"]).eachFileRecurse { file ->
            if (file.isFile() && file.name.endsWith(".groovy") && !(file.absolutePath =~ /junit/)) {
                String relPath = emi.canonicalPath(file) - emi.frameworkDir.absolutePath
                try {
                    Class.forName(file.name - ".groovy")
                    out?.debug "\t$relPath parsed successfully"
                } catch (CompilationFailedException cex) {
                    out?.error "The script [$relPath] failed to compile"
                    results.errors << "The script [$relPath] failed to compile"
                }
                catch (Exception e) {
                    out?.error "The script [$relPath] failed to compile"
                    results.errors << "The script [$relPath] failed to compile"
                }
            }
        }

        if (emi.frameworkInitialized) {
            // Confirm data mart and SAS Macro globals are configured
            // Only run if the framework has been initialized

            // Program list for BATCH Verification
            List sasPgmList = ['verifyProperties']
            List autoExecFiles = ['EVDMAUTOEXEC_BATCH.sas', 'EVDMAUTOEXEC_STP.sas']
            autoExecFiles.each { fName ->
                File file = emi.canonicalFile(emi.sysProps['sasev.datamartDir'], fName)
                if ( !file.canRead() ) {
                    out?.error "Missing or cannot read autoexec file [${file}]"
                    results.errors << "Missing or cannot read autoexec file [${file}]"
                    return results
                }
            }

            File autoExec = emi.canonicalFile(emi.sysProps['sasev.datamartDir'], 'EVDMAUTOEXEC_BATCH.sas')
            sasPgmList.each { p ->
                Map argsMap = [autoexec:autoExec.path]
                argsMap << [print:emi.canonicalPath(emi.sysProps['emi.log.dir'],"${p}_${runID}.lst")]

                if (verbose) {
                    argsMap << [verbose:verbose]
                    out?.info "Verifying Service Architecture properties"
                }

                File sasPgm = emi.canonicalFile(emi.sysProps['emi.framework.sasjobs.dir'], p + '.sas')

                if (!sasPgm.exists()) {
                    out?.error "Cannot find SAS program [${sasPgm}]"
                    results.errors << "Cannot find SAS program [${sasPgm}]"
                }
                int rc = emi.sas.run argsMap, sasPgm
                if (rc) {
                    String sasLog = "${p}_${runID}.sas.log"
                    out?.error "$sasPgm.name failed with return code [${rc}]. Check $sasLog for details."
                    results.errors << "$sasPgm.name failed with return code [${rc}]. Check $sasLog for details."
                    if (verbose) {
                        File sasLogFile = emi.canonicalFile(emi.logFile.parentFile, sasLog)
                        if (sasLogFile.exists()) {
                            out?.error "Content of $sasLog:"
                            out?.error sasLogFile.text
                        } else {
                            out?.error "$sasLogFile does not exist. Check ${emi.logFile} for details"
                        }
                    }
                } else {
                    if (verbose) { out?.info "\t[ran] [${p}] rc = 0" }
                }
            }
        } else {
            if (emi.acmInitialized) {
                results.errors << "Service Architecture is not initialized, but ACM ETL is initialized"
            }
            // apm_init can run before emi_init (at least for now)
            //if (emi.apmInitialized) {
            //    results.errors << "Service Architecture is not initialized, but APM ETL is initialized"
            //}
            if (emi.acmEnabled) {
                results.errors << "Service Architecture is not initialized, but ACM ETL is enabled"
            }
            if (emi.apmEnabled) {
                results.errors << "Service Architecture is not initialized, but APM ETL is enabled"
            }
        }

        out?.info "Level 2 validation completed with ${results.errors.size()} error(s) in ${(System.currentTimeMillis()-start)/1000.0} sec"
        return results
    }

    private ValidateResults validateLevel3() {
        ValidateResults results = new ValidateResults()

        out?.info "Starting level 3 validation..."
        long start = System.currentTimeMillis()

        if (emi.frameworkInitialized) {
            // Validate Reporting Properties
            List sasPgmList = []
            File autoExec
            sasPgmList = ['verifyReportingProps']
            autoExec = emi.canonicalFile(emi.sysProps['sasev.datamartDir'],'EVDMAUTOEXEC_STP.sas')
            if (!autoExec.canRead()) {
                results.errors << "Missing or cannot read autoexec file [${autoExec}]"
                out?.error "Missing or cannot read autoexec file [${autoExec}]"
                return results.errors
            }

            sasPgmList.each { p ->
                Map argsMap = [:]
                argsMap = [autoexec:autoExec.path]
                argsMap << [print:emi.canonicalPath(emi.sysProps['emi.log.dir'],"${p}_${runID}.lst")]

                if (verbose) {
                    argsMap << [verbose:verbose]
                    out?.info "Verifying Service Architecture reporting properties..."
                }

                File sasPgm = emi.canonicalFile(emi.sysProps['emi.framework.sasjobs.dir'], p + '.sas')

                if (!sasPgm.exists()) {
                    out?.error "Cannot find SAS program [${sasPgm}]"
                    results.errors << "Cannot find SAS program [${sasPgm}]"
                }
                int rc = emi.sas.run argsMap, sasPgm
                if (rc) {
                    String sasLog = "${p}_${runID}.sas.log"
                    out?.error "$sasPgm.name failed with return code [${rc}]. Check $sasLog for details."
                    results.errors << "$sasPgm.name failed with return code [${rc}]. Check $sasLog for details."
                    if (verbose) {
                        File sasLogFile = emi.canonicalFile(emi.logFile.parentFile, sasLog)
                        if (sasLogFile.exists()) {
                            out?.error "Content of $sasLog:"
                            out?.error sasLogFile.text
                        } else {
                            out?.error "$sasLogFile does not exist. Check ${emi.logFile} for details"
                        }
                    }
                } else {
                    if (verbose) { out?.info "\t[ran] [${p}] rc = 0" }
                }
            } // end of program list for batch STP Verification

            // Validate ETLs
            try {
                if (emi.acmEnabled) {
                    if (verbose) { out?.info "Checking ACM ETL..." }
                    results << checkETL("ACM", emi.canonicalFile(emi.frameworkDir, "Logs/metrics/acm_etl.metrics"))
                } else {
                    if (verbose) { out?.info "The ACM ETL is not enabled.  Skipping check." }
                }

                if (emi.apmEnabled) {
                    if (verbose) { out?.info "Checking APM ETL..." }
                    results << checkETL("APM", emi.canonicalFile(emi.frameworkDir, "Logs/metrics/apm_status.metrics"))
                } else {
                    if (verbose) { out?.info "The APM ETL is not enabled.  Skipping check." }
                }
            } catch (Exception e) {
                results.errors << "There was a problem validating ETL status: $e.message"
            }
        } else {
            results.warnings << "Level 3 validation requires Service Architecture be initialized"
        }

        out?.info "Level 3 validation completed with ${results.errors.size()} error(s) in ${(System.currentTimeMillis()-start)/1000.0} sec"
        return results
    }

    private ValidateResults validateLevel4() {
        ValidateResults results = new ValidateResults()

        out?.info "Starting level 4 validation..."
        long start = System.currentTimeMillis()

        // Validate checksums
        if (verbose) { out?.info "Validating file checksums..." }
        Checksum cs = new Checksum(emi)
        cs.out = emi.log
        if (cs.checksumZipFile.exists()) {
            Map checksumResults = cs.verify()
            if (checksumResults.modified) {
                checksumResults.modified.each {
                    results.warnings << "File modified: ${it - emi.frameworkDir - '/'}"
                }
            }
            if (checksumResults.missing) {
                checksumResults.missing.each {
                    results.errors <<  "File missing: ${it - emi.frameworkDir - '/'}"
                }
            }
            if (checksumResults.err) {
                checksumResults.err.each {
                    results.errors << "$it"
                }
            }
        } else {
            results.errors << "The Service Architecture checksum zip file ${cs.checksumZipFile.name} does not exist"
        }

        // Validate file owners
        if (verbose) { out?.info "Validating file ownership..." }
        results << validateFileOwners(emi.frameworkDir)

        // Write diagnostics
        if (verbose) { out?.info "Writing diagnostics..." }
        File diag = emi.canonicalFile(emi.frameworkDir, "Logs/diagnostics.txt")
        diag.delete()
        diag.createNewFile()
        out?.debug "Writing diagnostics..."
        results << writeDiagnostics(diag)
        if (verbose) { out?.info "Diagnostic information written to $diag" }
        out?.debug "Successfully wrote diagnostic information to $diag"

        out?.info "Level 4 validation completed with ${results.errors.size()} error(s) in ${(System.currentTimeMillis()-start)/1000.0} sec"
        return results
    }

    private ValidateResults checkETL(String etlName, File metricsFile) {
        ValidateResults results = new ValidateResults()

        Properties metrics = emi.readProperties(metricsFile)
        if (metrics["lastStartTime"]) {
            if (metrics["lastEndTime"]) {
                if (metrics["errorCount"] && ((metrics["errorCount"] as int) > 0)) {
                    results.errors << "The most recent execution of the $etlName ETL completed with ${metrics['errorCount']} error(s)"
                } else {
                    if (verbose) { out?.info "The $etlName ETL is operational" }
                }
            } else {
                results.errors << "The $etlName ETL is currently running or failed to exit properly"
            }
        } else {
            results.errors << "The $etlName ETL has not been started"
        }

        return results
    }

    // Check that most files under emi-framework are owned by the SAS install user
    ValidateResults validateFileOwners(File startDir) {
        ValidateResults results = new ValidateResults()

        List pathsToIgnore = [/LandingZone/, /ApmArchiveLog/, /Archive/, /Reports/]
        if (startDir?.exists() && startDir?.isDirectory()) {
            startDir.eachFileRecurse { file ->
                if (!(pathsToIgnore.any { file.absolutePath =~ it })) {
                    String fileOwner = Files.getOwner(file.toPath())?.name
                    if (emi.installUsername != fileOwner) {
                        String relPath = file.absolutePath - emi.frameworkDir.absolutePath - "/"
                        if (userPolicy == UserPolicy.ERROR) {
                            results.errors << "[$emi.installUsername] is not the owner of $relPath [$fileOwner]"
                        } else if (userPolicy == UserPolicy.WARN) {
                            out?.warn "WARNING: [$emi.installUsername] is not the owner of $relPath [$fileOwner]"
                        } else {
                            out?.debug "[$emi.installUsername] is not the owner of $relPath [$fileOwner] but the user policy is set to IGNORE"
                        }
                    }
                }
            }
        }

        return results
    }

    def writeDiagnostics(File diag) {
        ValidateResults results = new ValidateResults()
        // Header
        diag << "Service Architecture Diagnostic Utility\n"
        diag << "=======================================\n\n"
        diag << "Date    : ${new Date()}\n"
        diag << "Host    : ${emi.sysProps['sasev.agent.name']}\n"
        diag << "Version : ${emi.sysProps['framework.version']}\n"
        diag << "Build ID: ${emi.sysProps['framework.build.id']}\n"
        diag << "\n"

        Properties fw = emi.readProperties(emi.canonicalFile(emi.frameworkDir, "sasev.properties"))
        diag << "Service Architecture Properties\n"
        diag << "-------------------------------\n"
        (fw.keySet() as TreeSet).each { key ->
            diag << "\t$key = ${fw[key]}\n"
        }
        diag << "\n"

        List metricList = ['acm_etl.metrics','apm_status.metrics','sasevdm.metrics','log_centralize.metrics']
        metricList.each { mfile ->
            diag << "Metric File: [$mfile]\n"
            diag << "====================================\n"
            File metricFile = emi.canonicalFile(emi.sysProps['emi.log.dir'], "/metrics/$mfile")
            if (!metricFile.exists()) {
                results.errors << "\tMissing metrics file: [$mfile]"
            } else {
                diag.append(metricFile.getText())
            }
            diag << "\n"
        }

        // System Properties
        diag << "System Properties\n"
        diag << "-----------------\n"
        (System.properties.keySet() as TreeSet).each { key ->
            diag << "\t$key = ${System.getProperty(key)}\n"
        }
        diag << "\n"

        // SAS Environment Manager
        Map sasevConn = emi.hq.connectionInfo

        diag << "SAS Environment Manager Server\n"
        diag << "-----------------------\n"
        diag << "Server: ${sasevConn.hostname}\n"
        diag << "Port  : ${sasevConn.port}\n"
        diag << "User  : ${sasevConn.user}\n"
        diag << "\n"

        Properties agent = emi.readProperties(emi.canonicalFile(emi.hq.agentDir, "conf/agent.properties"))
        diag << "SAS Environment Manager Agent\n"
        diag << "-----------------------\n"
        diag << "Location: ${emi.hq.agentDir}\n"
        diag << "Version : ${emi.hq.agentDir.name - 'agent-' - '-EE'}\n"
        diag << "\n"
        return results
    }
}
