/*****************************************************************************
 * Copyright © 2015, SAS Institute Inc., Cary, NC, USA.  All Rights Reserved.
 ****************************************************************************/

// NOTE: Requires Service Architecture runtime and will not function standalone

// Master ETL script for APM.  This script coordinates running the individual
// APM ETL steps.

@BaseScript(MasterETLBaseScript)
import groovy.transform.BaseScript
import groovy.io.FileType
import groovy.json.JsonSlurper

import java.text.SimpleDateFormat
import java.io.File

// Globals
scratchDir = null

def pgmName = this.class.name
String runID = _LOG_TIMESTAMP

metricsFile = canonicalFile(sysProps["emi.log.dir"], "metrics/apm_status.metrics")
altLogDir = canonicalFile(sysProps["emi.log.dir"], "kits/emi-framework/Logs/")

// Set up expected CLI syntax
def cli = new CliBuilder(
    usage:  "master_apm_etl${scriptExtension} [-xvdh] [--evdebug <0|1|2|3>]",
    header: "Executes the ETL process for the APM component\n" +
            "Options:",
    footer: "Exit status:\n" +
            "$RC_SUCCESS  if OK\n" +
            "$RC_WARNING  if minor issues are encountered\n" +
            "$RC_GENERAL_ERROR  if an error is encountered\n" +
            "$RC_INVALID_USAGE for invalid usage"
)
cli.with {
    x(longOpt: 'noexit', 'Do not exit the JVM on completion', required: false)
    _(longOpt: 'evdebug', 'Set debug level for SAS invocations', args:1, type: int, argName: '0-3', required: false)
    v(longOpt:'verbose', 'Enable verbose output', required: false)
    d(longOpt:'debug', 'Enable debug output', required: false)
    h(longOpt:'help', 'Show usage information', required: false)
}

// Parse given arguments
OptionAccessor opts = cli.parse(args)
if (!opts) {
    return exitETL(RC_INVALID_USAGE)
}
if (opts.h) {
    cli.usage()
    return exitETL(RC_SUCCESS, false)
}
if (opts.v) {
    verbose = true
}
if (opts.d) {
    setLogLevel LogLevel.DEBUG
    evDebug = 2
    debug = true
}
if (opts.x) {
    exitJVM = false
}
if (opts.evdebug) {
    evDebug = opts.evdebug as int
}

// Property setting to enable and trigger the runReports capability ( only the appropriate report
// report bundle is selected) via the apm etl. run results method
boolean runReports = false
if (sysProps['custom.etl.runReports'] ==~ /(?i)true/) {
   runReports = true
}

// Verify Framework is initialized and APM is enabled
if (!frameworkInitialized) {
    msg.error "Service Architecture is not initialized.  The APM ETL cannot be run."
    errorCount++
    return exitETL(RC_GENERAL_ERROR, false)
}

msg.info "Service Architecture APM ETL Process - ${new Date()}"

if (isRunning("${this.class.name}" + ".groovy")) {
    msg.error "The APM ETL is already running. The existing process must complete before continuing."
    errorCount++
    return RC_GENERAL_ERROR
}

// Run Validator level 2
Validator v = new Validator(this, [verbose:verbose], log)
ValidateResults vResults = v.validate(2)
if (!vResults) {
    msg.error "Service Architecture validation failed. ${pgmName} cannot continue. Errors are:"
    vResults.errors.each { msg.error "\t$it" }
    msg.error "Stopping execution"
    return exitETL(RC_GENERAL_ERROR)
}

if (!apmEnabled) {
    msg.error "APM is not enabled.  The APM ETL cannot be run."
    errorCount++
    return exitETL(RC_WARNING, false)
}

String deployid = sysProps["sasev.deployment.id"] ?: net.fqdn.replaceAll(/\./,'_')

// Copy data to VA Autoload dropzone
boolean copy2VA = sysProps['vafeed.enabled'] ==~ /(?i)true/

// Export of APM data is currently not supported
// boolean exportData = sysProps['datamart.exportWithETL'] ==~ /(?i)true/
boolean exportData = false

// Create scratchDir if needed
evdmTempDir = canonicalFile(datamartDir, "evdmTemp")
scratchDir = canonicalFile(evdmTempDir, "scratch_${runID}")

if (!scratchDir.mkdir()) {
     msg.error "Cannot create [$scratchDir]. Exiting."
     return exitETL(RC_GENERAL_ERROR)
}

// Check that the working volumes have sufficient available space
// (defaults to 1000MB if not specified in sasev.properties)

if (!checkETLSpace()) {
     msg.error "Required minimum ETL working disk space not met. Exiting."
     return exitETL(RC_GENERAL_ERROR)
}


// Set up Environment Variables that will be passed to SAS
File sasFoundationDir = canonicalFile(sysProps["sas.root"])
apmEnv = [
    SASROOT:             canonicalPath(sasFoundationDir),
    CLASSPATH :          canonicalPath("$sasFoundationDir/misc/applets"),
    AGENTROOTDIR:        canonicalPath(sysProps["sasev.agent.dir"]),
    EVDEBUG:             "$evDebug"
]

// Options passed to all SAS invocations
commonSASOptions.logDir     = logDir.absolutePath
commonSASOptions.env        = apmEnv + [EVDEBUG:"$evDebug"] + [DEPLOYID:"$deployid"]
// commonSASOptions.workingDir = apmDir.absolutePath
commonSASOptions.verbose    = verbose
commonSASOptions.args       = COMMON_SAS_ARGS

// Make sure the landing zone directory for altlogs exists
altLogDir.mkdirs()

// Main body of ETL
try {
    statusInit()

    getTimezones()

    extractLandingZone()

    logProcess()

    runSASReports(runReports)

    if (exportData) {
       msg.info "Exporting Data."
       exportSomeData()
    }

    archiveLogs(".staging")

    if (copy2VA) {
       vaCopy()
    }
    
} catch (Exception e) {
    term.error "There was a problem executing the APM ETL: ${e.message}"
    log.error "There was a problem executing the APM ETL", e
    msg.error "Check $logFile for details"
    errorCount++
    return exitETL(RC_GENERAL_ERROR)
}

// Writing info event record
events.info "$pgmName executed"

return exitETL(errorCount > 0? RC_GENERAL_ERROR : RC_SUCCESS)
// =====================================================================
// End of Script
// =====================================================================

// =====================================================================
// Step Implementations
// =====================================================================

def checkMinSpace(String myprop, long defaultMinMB = 1000) {

// 'myprop' is a 'sysProps' property string representing a directory path.
// This routine will append ".MinSizeMB" to the supplied 'sysProps' property string
// and will use that to fetch the associated minimum required disk space for that path (in MB),
// comparing the usable space at that path to the supplied (or default) minimum amount.
// Returns 'false' if minimum value not met.

    long mySpaceMB

    String myminprop = myprop + '.MinSizeMB'

    long myminMB = Long.valueOf(sysProps[myminprop] ? sysProps[myminprop] : defaultMinMB)
    File myFile = canonicalFile(sysProps[myprop])

    try {
        mySpaceMB = myFile.getUsableSpace() / (1000 * 1000)
    } catch (Exception e) {
        term.error "Could not fetch usable space for $myFile: $e"
        log.error "Could not fetch usable space for $myFile", e
    }

    msg.debug "checkMinSpace: myprop is $myprop"
    msg.debug "checkMinSpace: Usable space in MB is $mySpaceMB, minimum limit is $myminMB\n"

    if (mySpaceMB < myminMB) {
        msg.error "Not enough disk space for ETL:"
        msg.error "   $myprop = $myFile"
        msg.error "   Usable space in MB is $mySpaceMB, required is $myminMB"
        return false
    }
    return true
}

def checkETLSpace() {
    space_OK = true

    if (!checkMinSpace('sasev.landingZoneDir')) {
        space_OK = false
        errorCount++
    }

    if (!checkMinSpace('sasev.archiveDir')) {
        space_OK = false
        errorCount++
    }

    if (!checkMinSpace('sasev.datamartDir')) {
        space_OK = false
        errorCount++
    }
    return space_OK
}

def statusInit() {
    // Write initial status
    boolean success = false
    try {
        if (!metricsFile.exists() || metricsFile.delete()) {
            writeProperty "lastStartTime", "${ISO8601_DATE.format(new Date())}", metricsFile
            success = true
        } else {
            msg.error "Unable to initialize APM status metrics file ${metricsFile}"
        }
    } catch (Exception e) {
        term.error "Unable to initialize APM status metrics: $e"
        log.error "Unable to initialize APM status metrics", e
    }

    if (!success) {
        errorCount++
        return exitETL(RC_GENERAL_ERROR)
    }
}

def getTimezones() {
    msg.info "Starting getTimezones..."
    long start = System.currentTimeMillis()

    File tzMetricsFile = canonicalFile(metricsDir, "apm_tz.metrics")
    initializeMetrics tzMetricsFile
    Map myOpts = [:] << commonSASOptions
    myOpts['env'] = acmEnv + apmEnv + [METRICSFILE:tzMetricsFile.absolutePath]
    myOpts['autoexec'] = canonicalPath(datamartDir, "EVDMAUTOEXEC_BATCH.sas")
    myOpts['logPrefix'] = "apm"
    dbMirrorData(tzMetricsFile, myOpts)
    inventoryUpdate(tzMetricsFile, myOpts)

    long duration = System.currentTimeMillis() - start
    (verbose? msg:log).info "getTimezones completed in $duration ms"
}


def logProcess() {
    msg.info "Starting log_process..."
    long start = System.currentTimeMillis()

    // =================
    // Artifact Analyzer
    // =================
    File sasProg = canonicalFile(jobsDir, "runLog_process.sas")

    Map myOpts = [:] << commonSASOptions
    myOpts["autoexec"] = canonicalPath(sysProps["sasev.datamartDir"], "EVDMAUTOEXEC_BATCH.sas")
    myOpts["args"] += [
        "-altlog", canonicalPath(altLogDir, "${sasProg.name-'.sas'}_${_LOG_TIMESTAMP}.sas.log"),
        "-memsize", "2g"
    ]

    msg.info "Executing ${sasProg.name}..."
    int rc = sas.run myOpts, sasProg
    if (rc > 1) {
        term.error "$sasProg.name execution failed with return code $rc"
        errorCount++
    }

    sasProg = canonicalFile(jobsDir, "runKit_parsers.sas")

    myOpts = [:] << commonSASOptions
    myOpts["autoexec"] = canonicalPath(sysProps["sasev.datamartDir"], "EVDMAUTOEXEC_BATCH.sas")
    myOpts["args"] += [
        "-altlog", canonicalPath(altLogDir, "${sasProg.name-'.sas'}_${_LOG_TIMESTAMP}.sas.log"),
        "-memsize", "2g"
    ]

    msg.info "Executing ${sasProg.name}..."
    rc = sas.run myOpts, sasProg
    if (rc > 1) {
        term.error "$sasProg.name execution failed with return code $rc"
        errorCount++
    }

    sasProg = canonicalFile(jobsDir, "runEMIProcessLogs.sas")

    // do not run with altlog assigned, job processes altlog
    myOpts = [:] << commonSASOptions
    myOpts["autoexec"] = canonicalPath(sysProps["sasev.datamartDir"], "EVDMAUTOEXEC_BATCH.sas")
    myOpts["args"] += [
        "-memsize", "2g"
    ]

    msg.info "Executing ${sasProg.name}..."
    rc = sas.run myOpts, sasProg
    if (rc > 1) {
        term.error "$sasProg.name execution failed with return code $rc"
        errorCount++
    } else {
        // altlog processing ran, deleting logs
        altLogDir.eachFile (FileType.FILES) {
        	if (it.name.endsWith('.log')) { it.delete() }
        }
    }


    long duration = System.currentTimeMillis() - start
    (verbose? msg:log).info "log_process completed in $duration ms"
}

def runSASReports(boolean runReports) {
    msg.info "Starting runSASReports..."
    long start = System.currentTimeMillis()

    File sasProg = canonicalFile(jobsDir, "MetadataGetArtifacts.sas")
    String[] sasArgs = [
        "-altlog", canonicalPath(altLogDir, "${sasProg.name-'.sas'}_${_LOG_TIMESTAMP}.sas.log"),
        "-memsize", "2g"
    ]

    Map myOpts = [:] << commonSASOptions
    myOpts["verbose"] = verbose
    myOpts["autoexec"] = canonicalPath(datamartDir, "EVDMAUTOEXEC_BATCH.sas")
    myOpts["args"] = sasArgs

    msg.info "Executing ${sasProg.name}"
    int rc = sas.run myOpts, sasProg
    if (rc > 1) {
        term.error "$sasProg.name execution failed with return code $rc"
        errorCount++
    }

    if (runReports) {

    	// run groovy program
    	msg.info "Running report bundles ARM and Audit"

		// call runReports need to support Audit, ARM, ACM, Kits
    	List pArgs = ['--rptBundle','ARM,Audit']
    	if (verbose) pArgs << '-v'
    	if (msg.debugEnabled) pArgs << '-d'
    	verbose? msg.info ("running batch report bundle ${pArgs}") : ""
    	rc = runGroovy(canonicalPath(sysProps['emi.framework.scripts.dir'],'runReports.groovy'), pArgs)
    	if (rc > 1) {
    		term.error "$sasProg.name execution failed with return code $rc"
        	errorCount++
    	}
    }

    long duration = System.currentTimeMillis() - start
    (verbose? msg:log).info "runSASReports completed in $duration ms"
}
def vaCopy() {
    msg.info "Starting vaCopy..."
    long start = System.currentTimeMillis()

    File vacMetricsFile = canonicalFile(metricsDir, "apm_vacopy.metrics")
    initializeMetrics vacMetricsFile
    Map myOpts = [:] << commonSASOptions
    myOpts['env'] = acmEnv + apmEnv + [METRICSFILE:vacMetricsFile.absolutePath] + [DATA2COPY2VA:"KIT2VA"]   

    myOpts['autoexec'] = canonicalPath(datamartDir, "EVDMAUTOEXEC_BATCH.sas")
    myOpts['logPrefix'] = "apm"
    copy2VA(vacMetricsFile, myOpts)

    long duration = System.currentTimeMillis() - start
    (verbose? msg:log).info "vaCopy completed in $duration ms"
}


// Returns the last non-empty (content other than whitespace) line in a file or null
String lastLine(File file) {
    String lastLine = null
    if (file!=null && file.isFile() && file.exists()) {
        try {
            file.eachLine { line ->
                if (line.trim()) {
                    lastLine = line
                }
            }
        } catch (IOException ioEx) {
            msg.warn "WARNING: Could not check file $file: $ioEx.message"
            lastLine = null
        }
    }

    return lastLine
}

def extractLandingZone() {
    msg.info "Extracting archives in Landing Zone..."

    AntBuilder ant = new AntBuilder()
    ant.project.getBuildListeners()[0].setMessageOutputLevel(verbose? 2 : 0)

    int extractedFileCount = 0
    File lzDir = canonicalFile(sysProps['sasev.landingZoneDir'])
    lzDir.mkdirs()
    def zipPattern = ~/.*\.zip/

    // Landing Zone contains zips of logs (The .staging subdirectory is temporary)
    File apmLz = canonicalFile(lzDir, ".staging")

    lzDir.eachFileMatch(zipPattern) { zipFile ->
        try {
            msg.info "\tExtracting ${zipFile.name}..."
            ant.unzip(src:"$zipFile", dest:"$apmLz", stripAbsolutePathSpec:"true")
            extractedFileCount++
            if (!zipFile.delete()) {
                msg.warn "Unable to delete ${zipFile.name}"
            }
        } catch (Exception e) {
            term.warn "Could not extract ${zipFile.name}. Processing will continue.: $e"
            log.warn "Could not extract ${zipFile.name}. Processing will continue.", e
        }
    }

    (verbose? msg:log).info "Landing Zone extraction completed processing of [$extractedFileCount] files"
}

// Archive logs in the given landing zone subdirectory to a zip file in the archive directory
int archiveLogs(String dirName) {
	int numErrors = 0
	File srcDir  = canonicalFile(sysProps['sasev.landingZoneDir'], dirName)
	File destDir = canonicalFile(sysProps['sasev.archiveDir'])

	if (srcDir.exists()) {
		msg.info "Archiving logs..."
	} else {
		(verbose? msg:log).info "No logs to archive.  $srcDir does not exist"
		return numErrors
	}

	AntBuilder ant = new AntBuilder()
	// Disable Ant output if not in debug mode
	if (!msg.debugEnabled && !verbose) {
		ant.project.buildListeners[0].messageOutputLevel = 0
	}
	try {
		File zipFile = canonicalFile(destDir, "logs_processed_${_LOG_TIMESTAMP}.zip")
		msg.debug "Creating archive ZIP file $zipFile.name from $srcDir..."
		ant.zip(
			destFile: zipFile.absolutePath,
			baseDir:  srcDir.absolutePath,
		)

		msg.debug "$zipFile.name created"

		if (errorCount == 0) {
			msg.debug "Deleting $srcDir..."
			ant.delete(
				dir: srcDir,
				verbose: "$verbose"
			)
			msg.debug "$srcDir deleted successfully"
		} else {
			msg.warn "The ETL has encountered errors and will not delete files in the Landing Zone " +
					 "so they can be processed again"
		}
	} catch (Exception e) {
		term.error "Archive failed: $e.message"
		log.error "Archive failed", e
		numErrors++
	}

	errorCount += numErrors
	return numErrors
}

def runStatusUpdate(int prevRC) {
    msg.debug "Starting runStatusUpdate..."
    long start = System.currentTimeMillis()

    // Library size metrics
    ['artifacts','status'].each { libName ->
        long sizeKB = dirSize(canonicalFile(datamartDir, "$libName"))/1024
        writeProperty "librarySize.$libName", "$sizeKB", metricsFile
    }

    writeProperty "lastEndTime", "${ISO8601_DATE.format(new Date())}", metricsFile

    writeProperty "returnCode", "$prevRC", metricsFile
    writeProperty "errorCount", "$errorCount", metricsFile

    long duration = System.currentTimeMillis() - start
    msg.debug "runStatusUpdate completed in $duration ms"
}

// Handle Exporting APM Data if necessary
int exportSomeData() {

    argsList = ['-x']
    if (verbose) argsList << '-v'
    if (msg.debugEnabled) argsList << '-d'
    argsList << '-apmlibmode'
    argsList << 'DELTAS'
    argsList << '-acmlibmode'
    argsList << 'NONE'
    argsList << '-runID'
    argsList << runID

    msg.info "Calling Data Mart Export script with options ${argsList}"
    rc = runGroovy canonicalPath(scriptDir,'exportDataMart.groovy'), argsList

    return rc

}

long dirSize(File dir) {
    long size = 0
    if (dir.exists()) {
		dir.eachFileRecurse(FileType.FILES) { file ->
			size += file.length()
		}
    }

    return size
}

// =====================================================================
// Helper Methods
// =====================================================================
int exitETL(int rc, boolean updateStatus=true) {
    if (updateStatus) {
        try {
            runStatusUpdate(rc)
        } catch (Exception e) {
            term.error "There was an error updating APM ETL status: $e"
            log.error "There was an error updating APM ETL status", e
        }

        long duration = System.currentTimeMillis() - masterStartTime
        term.info "Check $logFile for messages"
        msg.info "APM Master ETL Script completed in ${duration/1000} sec at ${new Date()} with $errorCount error(s)"
    }

    if (scratchDir)  {
        if (scratchDir.exists())  {
            if (scratchDir.deleteDir()) {
                msg.info "Deleting scratch area"
            } else {
                msg.error "Cannot delete [$scratchDir]"
                rc = RC_GENERAL_ERROR
            }
        }
    }

    if (exitJVM) {
        System.exit rc
    } else {
        return rc
    }
}
